from typing import Any, List, Optional

import apibara.starknet.proto.filter_pb2 as proto
from apibara.starknet.proto.types_pb2 import FieldElement


class Filter:
    def __init__(self):
        self._inner = proto.Filter()

    def encode(self) -> bytes:
        return self._inner.SerializeToString()

    def to_proto(self) -> proto.Filter:
        return self._inner

    def parse(self, raw: bytes):
        self._inner.ParseFromString(raw)

    def merge(self, other: "Filter") -> "Filter":
        """
        Returns a new filter that will generate data for both filters.

        Parameters
        ----------
        other : Filter
            the other filter
        """
        new_filter = Filter()

        # decide wether to include headers or not
        has_header = False
        has_weak_header = False
        if self._inner.HasField("header"):
            has_header = True
            if self._inner.header.weak:
                has_weak_header = True
        if other._inner.HasField("header"):
            has_header = True
            if other._inner.header.weak:
                has_weak_header = True
        if has_header:
            new_filter.with_header(weak=has_weak_header)

        # concat all other filters.
        new_filter._inner.transactions.extend(self._inner.transactions)
        new_filter._inner.transactions.extend(other._inner.transactions)

        new_filter._inner.events.extend(self._inner.events)
        new_filter._inner.events.extend(other._inner.events)

        new_filter._inner.messages.extend(self._inner.messages)
        new_filter._inner.messages.extend(other._inner.messages)

        # merge state update filters
        if self._inner.HasField("state_update") or other._inner.HasField(
            "state_update"
        ):
            raise RuntimeError(
                "merging state updates is not supported yet. Please open an issue on github."
            )

        return new_filter

    def with_header(self, weak: Optional[bool] = None) -> "Filter":
        """
        Include header in the returned data.

        Parameters
        ----------
        weak : bool, optional
            if True, only include header if any other filter matches.
        """
        self._inner.header.CopyFrom(proto.HeaderFilter(weak=weak))
        return self

    def add_transaction(self, tx: "TransactionFilter") -> "Filter":
        """
        Include transaction data. Use an empty filter to return all transactions.
        """
        self._inner.transactions.extend([tx.encode()])
        return self

    def add_event(self, ev: "EventFilter") -> "Filter":
        """
        Include event data. Use an empty filter to include all events.
        """
        self._inner.events.extend([ev.encode()])
        return self

    def add_message(self, msg: "MessageFilter") -> "Filter":
        """
        Include messages from L2 to L1. Use an empty filter to include all messages.
        """
        self._inner.messages.extend([msg.encode()])
        return self

    def with_state_update(self, state_update: "StateUpdateFilter") -> "Filter":
        """
        Include state updates.
        """
        self._inner.state_update.CopyFrom(state_update.encode())
        return self


class TransactionFilter:
    @classmethod
    def any(cls) -> "TransactionFilter":
        """
        Include any transaction.
        """
        return TransactionFilter()

    @classmethod
    def invoke_v0(cls) -> "InvokeV0TransactionFilter":
        """
        Include invoke transactions, v0.
        """
        return InvokeV0TransactionFilter()

    @classmethod
    def invoke_v1(cls) -> "InvokeV1TransactionFilter":
        """
        Include invoke transactions, v1.
        """
        return InvokeV1TransactionFilter()

    @classmethod
    def deploy(cls) -> "DeployTransactionFilter":
        """
        Include deploy transactions.
        """
        return DeployTransactionFilter()

    @classmethod
    def declare(cls) -> "DeclareTransactionFilter":
        """
        Include declare transactions.
        """
        return DeclareTransactionFilter()

    @classmethod
    def l1_handler(cls) -> "L1HandlerTransactionFilter":
        """
        Include L1 handle transactions.
        """
        return L1HandlerTransactionFilter()

    @classmethod
    def deploy_account(cls) -> "DeployAccountTransactionFilter":
        """
        Include deploy account transactions.
        """
        return DeployAccountTransactionFilter()

    def encode(self) -> proto.TransactionFilter:
        return proto.TransactionFilter()


class InvokeV0TransactionFilter:
    def __init__(self) -> None:
        self._inner = proto.InvokeTransactionV0Filter()

    def with_contract_address(
        self, address: FieldElement
    ) -> "InvokeV0TransactionFilter":
        """
        Filter by contract address.
        """
        self._inner.contract_address.CopyFrom(address)
        return self

    def with_entry_point_selector(
        self, selector: FieldElement
    ) -> "InvokeV0TransactionFilter":
        """
        Filter by entry point selector.
        """
        self._inner.entry_point_selector.CopyFrom(selector)
        return self

    def with_calldata(
        self, calldata: List[FieldElement]
    ) -> "InvokeV0TransactionFilter":
        """
        Filter by calldata prefix.
        """
        del self._inner.calldata[:]
        self._inner.calldata.extend(calldata)
        return self

    def encode(self) -> proto.TransactionFilter:
        return proto.TransactionFilter(invoke_v0=self._inner)


class InvokeV1TransactionFilter:
    def __init__(self) -> None:
        self._inner = proto.InvokeTransactionV1Filter()

    def with_sender_address(self, address: FieldElement) -> "InvokeV1TransactionFilter":
        """
        Filter by sender address.
        """
        self._inner.sender_address.CopyFrom(address)
        return self

    def with_calldata(
        self, calldata: List[FieldElement]
    ) -> "InvokeV1TransactionFilter":
        """
        Filter by calldata prefix.
        """
        del self._inner.calldata[:]
        self._inner.calldata.extend(calldata)
        return self

    def encode(self) -> proto.TransactionFilter:
        return proto.TransactionFilter(invoke_v1=self._inner)


class DeployTransactionFilter:
    def __init__(self) -> None:
        self._inner = proto.DeployTransactionFilter()

    def with_contract_address_salt(
        self, address: FieldElement
    ) -> "DeployTransactionFilter":
        """
        Filter by contract address salt.
        """
        self._inner.contract_address_salt.CopyFrom(address)
        return self

    def with_class_hash(self, class_hash: FieldElement) -> "DeployTransactionFilter":
        """
        Filter by class hash.
        """
        self._inner.class_hash.CopyFrom(class_hash)
        return self

    def with_constructor_calldata(
        self, calldata: List[FieldElement]
    ) -> "DeployTransactionFilter":
        """
        Filter by constructor calldata prefix.
        """
        del self._inner.constructor_calldata[:]
        self._inner.constructor_calldata.extend(calldata)
        return self

    def encode(self) -> proto.TransactionFilter:
        return proto.TransactionFilter(deploy=self._inner)


class DeclareTransactionFilter:
    def __init__(self) -> None:
        self._inner = proto.DeclareTransactionFilter()

    def with_sender_address(self, address: FieldElement) -> "DeclareTransactionFilter":
        """
        Filter by sender address.
        """
        self._inner.sender_address.CopyFrom(address)
        return self

    def with_class_hash(self, class_hash: FieldElement) -> "DeclareTransactionFilter":
        """
        Filter by class hash.
        """
        self._inner.class_hash.CopyFrom(class_hash)
        return self

    def encode(self) -> proto.TransactionFilter:
        return proto.TransactionFilter(declare=self._inner)


class L1HandlerTransactionFilter:
    def __init__(self) -> None:
        self._inner = proto.L1HandlerTransactionFilter()

    def with_contract_address(
        self, address: FieldElement
    ) -> "L1HandlerTransactionFilter":
        """
        Filter by contract address.
        """
        self._inner.contract_address.CopyFrom(address)
        return self

    def with_entry_point_selector(
        self, selector: FieldElement
    ) -> "L1HandlerTransactionFilter":
        """
        Filter by entry point selector.
        """
        self._inner.entry_point_selector.CopyFrom(selector)
        return self

    def with_calldata(
        self, calldata: List[FieldElement]
    ) -> "L1HandlerTransactionFilter":
        """
        Filter by calldata prefix.
        """
        del self._inner.calldata[:]
        self._inner.calldata.extend(calldata)
        return self

    def encode(self) -> proto.TransactionFilter:
        return proto.TransactionFilter(l1_handler=self._inner)


class DeployAccountTransactionFilter:
    def __init__(self) -> None:
        self._inner = proto.DeployAccountTransactionFilter()

    def with_contract_address_salt(
        self, address: FieldElement
    ) -> "DeployAccountTransactionFilter":
        """
        Filter by contract address salt.
        """
        self._inner.contract_address_salt.CopyFrom(address)
        return self

    def with_class_hash(
        self, class_hash: FieldElement
    ) -> "DeployAccountTransactionFilter":
        """
        Filter by class hash.
        """
        self._inner.class_hash.CopyFrom(class_hash)
        return self

    def with_constructor_calldata(
        self, calldata: List[FieldElement]
    ) -> "DeployAccountTransactionFilter":
        """
        Filter by constructor calldata prefix.
        """
        del self._inner.constructor_calldata[:]
        self._inner.constructor_calldata.extend(calldata)
        return self

    def encode(self) -> proto.TransactionFilter:
        return proto.TransactionFilter(deploy_account=self._inner)


class EventFilter:
    def __init__(self):
        self._inner = proto.EventFilter()

    def with_from_address(self, address: FieldElement) -> "EventFilter":
        """
        Filter by address of the contract emitting the event.
        """
        self._inner.from_address.CopyFrom(address)
        return self

    def with_keys(self, keys: List[FieldElement]) -> "EventFilter":
        """
        Filter by keys prefix.
        """
        del self._inner.keys[:]
        self._inner.keys.extend(keys)
        return self

    def with_data(self, data: List[FieldElement]) -> "EventFilter":
        """
        Filter by data prefix.
        """
        del self._inner.data[:]
        self._inner.data.extend(data)
        return self

    def encode(self) -> proto.EventFilter:
        return self._inner


class MessageFilter:
    def __init__(self) -> None:
        self._inner = proto.L2ToL1MessageFilter()

    def with_to_address(self, address: FieldElement) -> "MessageFilter":
        """
        Filter by address.
        """
        self._inner.to_address.CopyFrom(address)
        return self

    def with_payload(self, payload: List[FieldElement]) -> "MessageFilter":
        """
        Filter by payload prefix.
        """
        del self._inner.payload[:]
        self._inner.payload.extend(payload)
        return self

    def encode(self) -> proto.L2ToL1MessageFilter:
        return self._inner


class StateUpdateFilter:
    def __init__(self):
        self._inner = proto.StateUpdateFilter()

    def add_storage_diff(self, diff: "StorageDiffFilter") -> "StateUpdateFilter":
        """
        Includes all storage changes that match the filter.
        """
        self._inner.storage_diffs.extend([diff.encode()])
        return self

    def add_declared_contract(
        self, decl: "DeclaredContractFilter"
    ) -> "StateUpdateFilter":
        """
        Includes all declared contracts that match the filter.
        """
        self._inner.declared_contracts.extend([decl.encode()])
        return self

    def add_deployed_contract(
        self, deploy: "DeployedContractFilter"
    ) -> "StateUpdateFilter":
        """
        Includes all deployed contracts that match the filter.
        """
        self._inner.deployed_contracts.extend([deploy.encode()])
        return self

    def add_nonce_update(self, update: "NonceUpdateFilter") -> "StateUpdateFilter":
        """
        Includes all nonce updates that match the filter.
        """
        self._inner.nonces.extend([update.encode()])
        return self

    def encode(self) -> proto.StateUpdateFilter:
        return self._inner


class StorageDiffFilter:
    def __init__(self):
        self._inner = proto.StorageDiffFilter()

    def with_contract_address(self, address: FieldElement) -> "StorageDiffFilter":
        """
        Filter by contract address.
        """
        self._inner.contract_address.CopyFrom(address)
        return self

    def encode(self) -> proto.StorageDiffFilter:
        return self._inner


class DeclaredContractFilter:
    def __init__(self):
        self._inner = proto.DeclaredContractFilter()

    def with_class_hash(self, class_hash: FieldElement) -> "DeclaredContractFilter":
        """
        Filter by class hash.
        """
        self._inner.class_hash.CopyFrom(class_hash)
        return self

    def encode(self) -> proto.DeclaredContractFilter:
        return self._inner


class DeployedContractFilter:
    def __init__(self):
        self._inner = proto.DeployedContractFilter()

    def with_contract_address(self, address: FieldElement) -> "DeployedContractFilter":
        """
        Filter by contract address.
        """
        self._inner.contract_address.CopyFrom(address)
        return self

    def with_class_hash(self, class_hash: FieldElement) -> "DeployedContractFilter":
        """
        Filter by class hash.
        """
        self._inner.class_hash.CopyFrom(class_hash)
        return self

    def encode(self) -> proto.DeployedContractFilter:
        return self._inner


class NonceUpdateFilter:
    def __init__(self):
        self._inner = proto.NonceUpdateFilter()

    def with_contract_address(self, address: FieldElement) -> "NonceUpdateFilter":
        """
        Filter by contract address.
        """
        self._inner.contract_address.CopyFrom(address)
        return self

    def with_nonce(self, nonce: FieldElement) -> "NonceUpdateFilter":
        """
        Filter by nonce.
        """
        self._inner.nonce.CopyFrom(nonce)
        return self

    def encode(self) -> proto.NonceUpdateFilter:
        return self._inner



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import apibara.protocol.proto.stream_pb2 as stream__pb2


class StreamStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StreamData = channel.stream_stream(
            "/apibara.node.v1alpha2.Stream/StreamData",
            request_serializer=stream__pb2.StreamDataRequest.SerializeToString,
            response_deserializer=stream__pb2.StreamDataResponse.FromString,
        )


class StreamServicer(object):
    """Missing associated documentation comment in .proto file."""

    def StreamData(self, request_iterator, context):
        """Stream data from the node."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_StreamServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "StreamData": grpc.stream_stream_rpc_method_handler(
            servicer.StreamData,
            request_deserializer=stream__pb2.StreamDataRequest.FromString,
            response_serializer=stream__pb2.StreamDataResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "apibara.node.v1alpha2.Stream", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Stream(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def StreamData(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            "/apibara.node.v1alpha2.Stream/StreamData",
            stream__pb2.StreamDataRequest.SerializeToString,
            stream__pb2.StreamDataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from typing import Optional, Union

import apibara.starknet.felt as felt
from apibara.protocol.proto.stream_pb2 import Cursor
from apibara.starknet.proto.types_pb2 import FieldElement


def starknet_cursor(
    block_number: int, block_hash: Optional[Union[bytes, str, FieldElement]] = None
) -> Cursor:
    """
    Create a new StarkNet cursor for the given block number and block hash.

    Arguments
    ---------
    block_number : int
        block height
    block_hash : bytes | str | FieldElement, optional
        block hash
    """
    if block_hash is not None:
        unique_key = _to_bytes(block_hash)
    else:
        unique_key = b""
    return Cursor(order_key=block_number, unique_key=unique_key)


def _to_bytes(hash: Union[bytes, str, FieldElement]) -> bytes:
    if isinstance(hash, bytes):
        return hash

    if isinstance(hash, FieldElement):
        hash = felt.to_hex(hash)

    if isinstance(hash, str):
        return bytes.fromhex(hash)

    raise ValueError("hash must be bytes, str, or FieldElement")



import apibara.starknet.felt as felt
from apibara.starknet.cursor import *
from apibara.starknet.filter import *
from apibara.starknet.indexer import *
from apibara.starknet.proto.starknet_pb2 import Block
from apibara.starknet.proto.types_pb2 import FieldElement



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: filter.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n\x0c\x66ilter.proto\x12\x19\x61pibara.starknet.v1alpha2\x1a\x0btypes.proto\"\xc3\x02\n\x06\x46ilter\x12\x37\n\x06header\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.HeaderFilter\x12\x42\n\x0ctransactions\x18\x02 \x03(\x0b\x32,.apibara.starknet.v1alpha2.TransactionFilter\x12\x42\n\x0cstate_update\x18\x03 \x01(\x0b\x32,.apibara.starknet.v1alpha2.StateUpdateFilter\x12\x36\n\x06\x65vents\x18\x04 \x03(\x0b\x32&.apibara.starknet.v1alpha2.EventFilter\x12@\n\x08messages\x18\x05 \x03(\x0b\x32..apibara.starknet.v1alpha2.L2ToL1MessageFilter\"\x1c\n\x0cHeaderFilter\x12\x0c\n\x04weak\x18\x01 \x01(\x08\"\xe3\x03\n\x11TransactionFilter\x12I\n\tinvoke_v0\x18\x01 \x01(\x0b\x32\x34.apibara.starknet.v1alpha2.InvokeTransactionV0FilterH\x00\x12I\n\tinvoke_v1\x18\x02 \x01(\x0b\x32\x34.apibara.starknet.v1alpha2.InvokeTransactionV1FilterH\x00\x12\x44\n\x06\x64\x65ploy\x18\x03 \x01(\x0b\x32\x32.apibara.starknet.v1alpha2.DeployTransactionFilterH\x00\x12\x46\n\x07\x64\x65\x63lare\x18\x04 \x01(\x0b\x32\x33.apibara.starknet.v1alpha2.DeclareTransactionFilterH\x00\x12K\n\nl1_handler\x18\x05 \x01(\x0b\x32\x35.apibara.starknet.v1alpha2.L1HandlerTransactionFilterH\x00\x12S\n\x0e\x64\x65ploy_account\x18\x06 \x01(\x0b\x32\x39.apibara.starknet.v1alpha2.DeployAccountTransactionFilterH\x00\x42\x08\n\x06\x66ilter\"\xe0\x01\n\x19InvokeTransactionV0Filter\x12\x41\n\x10\x63ontract_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x45\n\x14\x65ntry_point_selector\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x39\n\x08\x63\x61lldata\x18\x03 \x03(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\x97\x01\n\x19InvokeTransactionV1Filter\x12?\n\x0esender_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x39\n\x08\x63\x61lldata\x18\x03 \x03(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\xe5\x01\n\x17\x44\x65ployTransactionFilter\x12\x46\n\x15\x63ontract_address_salt\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12;\n\nclass_hash\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x45\n\x14\x63onstructor_calldata\x18\x04 \x03(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\x98\x01\n\x18\x44\x65\x63lareTransactionFilter\x12;\n\nclass_hash\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12?\n\x0esender_address\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\xe1\x01\n\x1aL1HandlerTransactionFilter\x12\x41\n\x10\x63ontract_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x45\n\x14\x65ntry_point_selector\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x39\n\x08\x63\x61lldata\x18\x03 \x03(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\xec\x01\n\x1e\x44\x65ployAccountTransactionFilter\x12\x46\n\x15\x63ontract_address_salt\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12;\n\nclass_hash\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x45\n\x14\x63onstructor_calldata\x18\x04 \x03(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\x8c\x01\n\x13L2ToL1MessageFilter\x12;\n\nto_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x38\n\x07payload\x18\x02 \x03(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\xba\x01\n\x0b\x45ventFilter\x12=\n\x0c\x66rom_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x35\n\x04keys\x18\x02 \x03(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x35\n\x04\x64\x61ta\x18\x03 \x03(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\xc8\x03\n\x11StateUpdateFilter\x12\x43\n\rstorage_diffs\x18\x01 \x03(\x0b\x32,.apibara.starknet.v1alpha2.StorageDiffFilter\x12M\n\x12\x64\x65\x63lared_contracts\x18\x02 \x03(\x0b\x32\x31.apibara.starknet.v1alpha2.DeclaredContractFilter\x12M\n\x12\x64\x65ployed_contracts\x18\x03 \x03(\x0b\x32\x31.apibara.starknet.v1alpha2.DeployedContractFilter\x12<\n\x06nonces\x18\x04 \x03(\x0b\x32,.apibara.starknet.v1alpha2.NonceUpdateFilter\x12H\n\x10\x64\x65\x63lared_classes\x18\x05 \x03(\x0b\x32..apibara.starknet.v1alpha2.DeclaredClassFilter\x12H\n\x10replaced_classes\x18\x06 \x03(\x0b\x32..apibara.starknet.v1alpha2.ReplacedClassFilter\"V\n\x11StorageDiffFilter\x12\x41\n\x10\x63ontract_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"U\n\x16\x44\x65\x63laredContractFilter\x12;\n\nclass_hash\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\x98\x01\n\x13\x44\x65\x63laredClassFilter\x12;\n\nclass_hash\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x44\n\x13\x63ompiled_class_hash\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\x95\x01\n\x13ReplacedClassFilter\x12\x41\n\x10\x63ontract_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12;\n\nclass_hash\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\x98\x01\n\x16\x44\x65ployedContractFilter\x12\x41\n\x10\x63ontract_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12;\n\nclass_hash\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\"\x8e\x01\n\x11NonceUpdateFilter\x12\x41\n\x10\x63ontract_address\x18\x01 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElement\x12\x36\n\x05nonce\x18\x02 \x01(\x0b\x32'.apibara.starknet.v1alpha2.FieldElementb\x06proto3"
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "filter_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _FILTER._serialized_start = 57
    _FILTER._serialized_end = 380
    _HEADERFILTER._serialized_start = 382
    _HEADERFILTER._serialized_end = 410
    _TRANSACTIONFILTER._serialized_start = 413
    _TRANSACTIONFILTER._serialized_end = 896
    _INVOKETRANSACTIONV0FILTER._serialized_start = 899
    _INVOKETRANSACTIONV0FILTER._serialized_end = 1123
    _INVOKETRANSACTIONV1FILTER._serialized_start = 1126
    _INVOKETRANSACTIONV1FILTER._serialized_end = 1277
    _DEPLOYTRANSACTIONFILTER._serialized_start = 1280
    _DEPLOYTRANSACTIONFILTER._serialized_end = 1509
    _DECLARETRANSACTIONFILTER._serialized_start = 1512
    _DECLARETRANSACTIONFILTER._serialized_end = 1664
    _L1HANDLERTRANSACTIONFILTER._serialized_start = 1667
    _L1HANDLERTRANSACTIONFILTER._serialized_end = 1892
    _DEPLOYACCOUNTTRANSACTIONFILTER._serialized_start = 1895
    _DEPLOYACCOUNTTRANSACTIONFILTER._serialized_end = 2131
    _L2TOL1MESSAGEFILTER._serialized_start = 2134
    _L2TOL1MESSAGEFILTER._serialized_end = 2274
    _EVENTFILTER._serialized_start = 2277
    _EVENTFILTER._serialized_end = 2463
    _STATEUPDATEFILTER._serialized_start = 2466
    _STATEUPDATEFILTER._serialized_end = 2922
    _STORAGEDIFFFILTER._serialized_start = 2924
    _STORAGEDIFFFILTER._serialized_end = 3010
    _DECLAREDCONTRACTFILTER._serialized_start = 3012
    _DECLAREDCONTRACTFILTER._serialized_end = 3097
    _DECLAREDCLASSFILTER._serialized_start = 3100
    _DECLAREDCLASSFILTER._serialized_end = 3252
    _REPLACEDCLASSFILTER._serialized_start = 3255
    _REPLACEDCLASSFILTER._serialized_end = 3404
    _DEPLOYEDCONTRACTFILTER._serialized_start = 3407
    _DEPLOYEDCONTRACTFILTER._serialized_end = 3559
    _NONCEUPDATEFILTER._serialized_start = 3562
    _NONCEUPDATEFILTER._serialized_end = 3704
# @@protoc_insertion_point(module_scope)



from apibara.indexer.indexer import Indexer
from apibara.starknet.filter import Filter
from apibara.starknet.proto.starknet_pb2 import Block


class StarkNetIndexer(Indexer[Filter, Block]):
    def encode_filter(self, filter: Filter) -> bytes:
        return filter.encode()

    def decode_data(self, raw: bytes) -> Block:
        block = Block()
        block.ParseFromString(raw)
        return block



import asyncio
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Generic, Optional

from grpc import StatusCode
from grpc.aio import AioRpcError

from apibara.indexer.configuration import Data, Filter, IndexerConfiguration
from apibara.indexer.info import Info, UserContext
from apibara.protocol.proto.stream_pb2 import Cursor


@dataclass
class Reconnect:
    """
    Reconnect configuration.

    Parameters
    ----------
    reconnect: bool
        reconnect to the stream
    cursor: Cursor
        reconnect from this cursor
    """

    reconnect: bool
    cursor: Optional[Cursor] = None


class Indexer(Generic[Filter, Data], metaclass=ABCMeta):
    def __init__(self):
        self._new_filter = None

    def update_filter(self, filter: Filter):
        """Updates the indexer filter by merging the current filter with the new one"""
        if self._new_filter is None:
            self._new_filter = filter
        else:
            self._new_filter = self._new_filter.merge(filter)

    def _get_and_reset_filter(self) -> Optional[Filter]:
        if self._new_filter is None:
            return None
        self._new_filter, filter = None, self._new_filter
        return filter

    @abstractmethod
    def indexer_id(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def initial_configuration(self) -> IndexerConfiguration[Filter]:
        raise NotImplementedError()

    @abstractmethod
    def encode_filter(self, filter: Filter) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    def decode_data(self, raw: bytes) -> Data:
        raise NotImplementedError()

    @abstractmethod
    async def handle_data(self, info: Info[UserContext, Filter], data: Data):
        raise NotImplementedError()

    async def handle_pending_data(self, info: Info[UserContext, Filter], data: Data):
        raise NotImplementedError()

    async def handle_invalidate(self, info: Info[UserContext, Filter], cursor: Cursor):
        return

    async def handle_reconnect(self, exc: Exception, retry_count: int) -> Reconnect:
        if not isinstance(exc, AioRpcError):
            return Reconnect(reconnect=False)

        if exc.code != StatusCode.INTERNAL:
            return Reconnect(reconnect=False)

        await asyncio.sleep(retry_count)
        return Reconnect(reconnect=retry_count < 5)



import asyncio
from asyncio.queues import Queue
from typing import AsyncIterable, Iterable, Optional, Tuple, Union

import grpc
from grpc.aio import Channel

from apibara.protocol.proto.stream_pb2 import (Cursor, DataFinality,
                                               StreamDataRequest,
                                               StreamDataResponse)
from apibara.protocol.proto.stream_pb2_grpc import StreamStub

DEFAULT_TIMEOUT = 45.0


class BearerTokenAuth(grpc.AuthMetadataPlugin):
    def __init__(self, token: str):
        if token is not None:
            self._meta = [("authorization", f"bearer {token}")]
        else:
            self._meta = []

    def __call__(self, context, callback):
        callback(self._meta, None)


def credentials_with_auth_token(token: str, credentials: grpc.CallCredentials):
    """Returns new credentials that authenticate with the server."""
    return grpc.composite_channel_credentials(
        credentials, grpc.metadata_call_credentials(BearerTokenAuth(token))
    )


class StreamService:
    """
    An Apibara Stream service.

    Arguments
    ---------
    channel: grpc.aio.Channel
        the grpc channel
    token: str, optional
        authentication token sent to the server
    """

    def __init__(self, channel: Channel) -> None:
        self._channel = channel
        self._stub = StreamStub(self._channel)

    def stream_data(self, timeout=None) -> Tuple["StreamClient", "StreamIter"]:
        """Start streaming data from the server.

        Arguments
        ---------
        timeout: float
            timeout in seconds for waiting messages from the server.
        """
        if timeout is None:
            timeout = DEFAULT_TIMEOUT
        ctx = _Context()
        client = StreamClient(ctx)
        iter = self._stub.StreamData(client.client_stream)
        stream = StreamIter(ctx, iter, timeout)
        return client, stream


class StreamClient:
    def __init__(self, ctx: "_Context") -> None:
        self._ctx = ctx
        self.client_stream = _RequestChannel()

    async def configure(
        self,
        *,
        filter: Optional[bytes] = None,
        batch_size: Optional[int] = None,
        finality: Optional[DataFinality.ValueType] = None,
        cursor: Optional[Cursor] = None,
    ):
        self._ctx.stream_id += 1
        request = StreamDataRequest(
            stream_id=self._ctx.stream_id,
            filter=filter,
            starting_cursor=cursor,
            batch_size=batch_size,
            finality=finality,
        )
        await self.client_stream.push(request)


class StreamIter:
    def __init__(
        self, ctx: "_Context", inner: grpc.StreamStreamMultiCallable, timeout: float
    ) -> None:
        self._ctx = ctx
        self._inner = inner
        self._timeout = timeout
        self._iter = None

    def __aiter__(self) -> AsyncIterable[StreamDataResponse]:
        self._iter = self._inner.__aiter__()
        return self

    async def __anext__(self):
        while True:
            value = await asyncio.wait_for(
                self._iter.__anext__(), timeout=self._timeout
            )
            if value.stream_id == self._ctx.stream_id:
                return value


class _Context:
    def __init__(self):
        self.stream_id = 0


class _RequestChannel:
    def __init__(self) -> None:
        self._q = Queue(maxsize=32)

    async def push(self, request: StreamDataRequest):
        await self._q.put(request)

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await self._q.get()
